import 'dart:async';

import 'package:flex/flex.dart';
import 'package:meta/meta.dart';
import 'package:plexus/plexus.dart';

/// Represents a class holding a message
abstract class IHasShortMessage {
  /// The short message for this event. Should be less than 40 characters.
  String get message;
}

/// Represents a class holding a timestamp
abstract class IHasTimestamp {
  /// The timestamp for this event.
  DateTime get timestamp;
}

/// Represents a class holding a long message
abstract class IHasLongMessage {
  /// The long message for this event. Should contain the same information as [IHasMessage.message] and then expand on it.
  @nullable
  String get longMessage;
}

/// Represents a class holding a source
abstract class IHasSource {
  /// The source of this event.
  Module get source;
}

/// Represents a class holding a blob of type [B]
abstract class IHasBlob<B> {
  /// Non-string message passed to Logger.
  @nullable
  B get blob;
}

/// Represents a class holding a level
abstract class IHasLevel {
  /// The level of this event.
  Level get level;
}

/// Represents a class holding an error and optional stacktrace
abstract class IHasError {
  /// Associated error (if any) when recording errors messages.
  @nullable
  Object get error;

  /// Associated stackTrace (if any) when recording errors messages.
  @nullable
  StackTrace get stackTrace;
}

/// Represents a class holding a zone
abstract class IHasZone {
  /// Zone of the calling code which resulted in this LogRecord.
  Zone get zone;
}

/// Represents a class holding a state of init or not init
abstract class IHasIsInit {
  /// Does this event take place during initialization
  bool get init;
}

/// An event which will be logged by a [Logger]'s individual handlers.
///
/// [source] will usually be set by a Logger.
/// Manually setting [timestamp] is discouraged.
@optionalTypeArgs
abstract class Event<B> implements IHasSource, IHasShortMessage, IHasTimestamp, IHasLevel, IHasZone {
  /// Should be called by subclass constructors.
  Event.i(@nullable DateTime timestamp, this.source, @nullable Zone zone) : timestamp = timestamp ?? DateTime.now(), zone = zone ?? Zone.current;

  /// Creates a new instance of [_Event].
  factory Event(Module source, Level level, String message, {@nullable String longMessage, @nullable Error error, @nullable StackTrace stackTrace, @nullable B blob, @nullable Zone zone, bool init = false, @nullable DateTime timestamp}) => StdEvent<B>(source, level, message, longMessage: longMessage, error: error, stackTrace: stackTrace, blob: blob, zone: zone, init: init, timestamp: timestamp);

  static StdEvent<String> parse(String encoded, Module source, @nullable Zone zone) {
    assert(encoded.startsWith(encodedEventHeader));

    final e = encoded.split(encodedEventSeparator);

    return StdEvent<String>(source, Level.encodedToPlexus[e[2]], e[3], blob: e.length == 5 ? e[4] : null);
  }

  @override
  final DateTime timestamp;

  @override
  final Module source;

  @override
  final Zone zone;
}

/// The default implementation of [Event]
class StdEvent<B> extends Event<B> implements IHasLongMessage, IHasBlob<B>, IHasError, IHasIsInit {
  /// Creates a new instance of [StdEvent]
  StdEvent(Module source, this.level, this.message, {this.longMessage, this.error, this.stackTrace, this.blob, @nullable Zone zone, this.init = false, @nullable DateTime timestamp})
      : super.i(timestamp, source, zone);

  @override
  final Level level;

  @override
  final String message;

  @override
  @nullable
  final String longMessage;

  @override
  @nullable
  final B blob;

  @override
  @nullable
  final Object error;

  @override
  @nullable
  final StackTrace stackTrace;

  @override
  final bool init;
}

/// An implementation of [Event] representing a change in authentication state
class AuthEvent<B> extends Event<B> implements IHasLongMessage, IHasBlob<B>, IHasError, IHasIsInit {
  /// Creates a new instance of [AuthEvent]
  AuthEvent(Module source, this.level, this.message, this.signal, {this.longMessage, this.error, this.stackTrace, this.blob, @nullable Zone zone, this.init = false, @nullable DateTime timestamp})
      : super.i(timestamp, source, zone);

  @override
  final Level level;

  @override
  final String message;

  @override
  @nullable
  final String longMessage;

  @override
  @nullable
  final B blob;

  @override
  @nullable
  final Object error;

  @override
  @nullable
  final StackTrace stackTrace;

  @override
  final bool init;

  /// The [AuthSignal] of this event.
  final AuthSignal signal;
}

/// An implementation of [Event] representing a request for user input
class RequestInputEvent<B> extends Event<B> {
  /// Creates a new instance of [RequestInputEvent]
  RequestInputEvent(Module source, this.context, this.message, this.test, {@nullable Zone zone, @nullable DateTime timestamp})
      : super.i(timestamp, source, zone);

  @override
  final Level level = Level.PRINT;

  /// This will be used for the input prompt.
  @override
  final String message;

  final List<String> context;

  final InputTest test;

  bool testInput(String current) => test(context, current);
}
